'' E1.31 to DMX bridge firmware v1.00 B6
'' By Robert P. Martin
'' Last revised 06/22/2011
'' Based on Jim St.John's E680 v2.06 E1.31 Pixel Driver firmware. http://www.sandevices.com
''
'' See "Terms Of Use" at end of file
''
'' Author's disclaimer:  The Web page logic is complex due to having to generate web pages while simultaneously processing
'' E1.31 data packets on the other sockets without dropping packets.
''
'' NOTE: This software REQUIRES a 256kB EEPROM because we store config params above 32k
''
'' Display network mode on startup by flashing the Green LED so you know whether it's STATIC or DHCP
'' 1= Normal DHCP  2= Normal STATIC 3= FORCED DHCP  4= FORCED STATIC 7= FORCED DHCP FAILED, default IP to 169.254.74.73
''
'' MODE button functionality in this version.
'' Press both RESET and MODE buttons, then release RESET while continuing to hold the MODE button.
'' The number of flashes of both Red and Green LED's while the MODE button is held indicates startup mode.
'' The function is selected by releasing the MODE button at the specified number of flashes.
''
'' 1 flash   = Normal Network, but force Webserver ON
'' 2 flashes = Force STATIC IP and Webserver ON
'' 3 flashes = Force DHCP and Webserver ON
''
'' Recovery modes -  To reset the interface back to defaults:
'' 8 flashes = Reset all settings to defaults and store in all 8 pages of EEPROM
'' 10 flashes= Reset all settings to defaults, generate new random MAC address (if no MAC IC present) and store in all 8 pages of EEPROM
''
''
'' Description:
'' This firmware will receive up to four universes of multicast E1.31 data
'' via a WIZ812 ethernet module and drive up to four DMX512 universe outputs,
'' or up to four RENARD universe outputs, or any combination thereof, depending
'' on configured output options.
''
'' On boot, MAC address is read from Microchip MAC PROM, if the 1st 3 bytes aren't Microchip OUI($00:$04:$A3)
'' we assume the IC is not present and we then check for our generated MAC prefix ($4A:$49:$4D).
'' If these aren't correct we will generate a new MAC using THE 1st 3 bytes of (4A,49,4D) and use a
'' random number generator routine to create the last 3 bytes, write the MAC to EEPROM, then reboot. 
''
'' There are 8 banks of EEPROM storage for system parameters.
'' Banks start at: bank1=$8000, bank2=$8040, bank3=$8080, bank4=$80C0
''                 bank5=$8100, bank6=$8140, bank7=$8180, bank8=$81C0
'' EEPROM addresses used by bank 1 (with an allocation of $40 bytes per bank).
'' 
'' $8000-$8007      Bank DMX offsets - 4 words
'' $8008-$800D      MAC address - 6 bytes
'' $800E-$8011      Bank protocols settings - 4 bytes
'' $8012-$8015      Static IP - 4 bytes
'' $8016-$8019      Static Gateway - 4 bytes 
'' $801A-$801D      Renard Baudrate settings - 4 bytes
'' $801E            Bootmode - 1 byte
'' $801F            Webmode - 1 byte
'' $8020            IPmode - 1 byte                                                                         
'' $8021            LOSmode - 1 byte
'' $8022-$803F      Unused - For future use

CON

   MajorVersion = 1             'Software Revision Level
   MinorVersion = 00              '
   
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  Second        = _xinfreq*16   'Delay for 1 sec
  _OUTPUT       = 1             'Sets pin to output in DIRA register
  _INPUT        = 0             'Sets pin to input in DIRA register  
  _HIGH         = 1             'High=ON=1=3.3v DC
  _LOW          = 0             'Low=OFF=0=0v DC
  _ON           = 1
  _OFF          = 0
  _ENABLE       = 1             'Enable (turn ON) function/mode
  _DISABLE      = 0             'Disable (turn OFF) function/mode

' Output pin definitions - Port 1:
  OUT1_Rx       = 0             'Receive pin for Port 1
  OUT1_Tx       = 1             'Transmit pin for Port 1
  OUT1_Ctrl     = 2             'Control pin; low for RX, high for TX
' Pin 4 used for ProgPin - see below

' Output pin definitions - Port 2:
  OUT2_Rx       = 3             'Receive pin for Port 2
  OUT2_Tx       = 4             'Transmit pin for Port 2
  OUT2_Ctrl     = 5             'Control pin; low for RX, high for TX

' Output pin definitions - Port 3:
  OUT3_Rx       = 8             'Receive pin for Port 3
  OUT3_Tx       = 9             'Transmit pin for Port 3
  OUT3_Ctrl     = 10            'Control pin; low for RX, high for TX

' Output pin definitions - Port 4:
  OUT4_Rx       = 13            'Receive pin for Port 4
  OUT4_Tx       = 14            'Transmit pin for Port 4
  OUT4_Ctrl     = 15            'Control pin; low for RX, high for TX

' DMX Parameters - All Ports
  DMX_b2bt      = 25            'Break-to-Break timing, 24ms to 44ms - Enter 0 for variable timing based on number of channels

' Renard Parameters - All Ports
  REN_Refresh   = 0
 
' LED's
  GrnLED        = 16            'Grn LED - high to light
  RedLED        = 17            'Red LED - high to light

' EEPROM definitions
  _EEPROMsel    = $a0           'Device select code for EEPROM for I2C routines
  _EEPROMbase   = $8000         'We will use EEPROM addresses starting at $8000 for our params
  _EEPROMpin    = 28            'SCL clock pin of EEPROM Clock, SDA data is assumed +1 pin

  _MACROMsel    = $a2           'Device select code for MAC ROM
  _MACROMbase   = $fa           'We will use MAC ROM address starting at $FA for our MAC read
  _MACROMpin    = 28            'SCL clock pin of MAC ROM Clock, SDA data is assumed +1 pin

' MODE pushbutton pin assignment
  ProgPin       = 18            'Input pin for MODE pushbutton
                                
' WIZnet module I/O pin definitions, using the SPI mode driver
  _RESET        = 20            'WIZ module I/O Reset
  _SCLK         = 19            'WIZ module I/O Serial Slock
  _SCS          = 21            'WIZ module I/O Serial Chip Select
  _MISO         = 22            'WIZ module I/O Master In Slave Out
  _MOSI         = 23            'WIZ module I/O Master Out Slave In

' Misc definitions
  MaxProtocols  = 4             'Max number of output Protocols supported 
  MaxTP         = 9             'Max defined Test Pattern# (not yet implemented)
  MaxBootmode   = 3             'Max defined Boot Mode value
  MaxLosMode    = 199           'Max defined Loss Of Signal mode (not yet implemented)
  MaxWebMode    = 3             'Max defined Web server modes
  MaxIPmode     = 1             'Max defined IP mode, currently defined are Static and DHCP

' Literals for web server
  LocalSocket   = 80            'Port for HTTP In/Out
  DestSocket    = 80            'Port for HTTP In/Out
  CR            = 13            'Carriage Return 
  LF            = 10            'Line Feed
  QUOTE         = 34            'Quotes
  E131PacketBufferLength = 648  'Buffer size for E1.31 packets
  WebPacketBufferLength  = 2048 'Buffer size for Web server incoming packets, prob could be quite a bit smaller
                                'but then would have to test for too-large packet in asn 1st

DAT

  DEFAULT_HOST_NAME         BYTE "E131_DMX_BRIDGE", 0

  RLPstr                    BYTE $00,$10,$00,$00,$41,$53,$43,$2d,$45,$31,$2e,$31,$37,$00
                            '1st check string to verify Root Layer is an E1.31 packet
                            'comprised of Preamble size /Postamble size/ACN Packet ID
  DMPstr                    BYTE $02,$A1,$00,$00,$00,$00,$01,$00
                            '2nd check string to verify DMP Layer is an E1.31 packet
                            
'These strings are used by the input command parser.
'The first two left characters in the line are parsed for the command, followed by the command parameters.
'The commands are:
   'BOot, PRotocol, DHcp, GAteway, IP, LOad, NOactivity
   'PAssword, QUit, SAve, SUbnet, TEst, UNiverse, WEbmode, DNs, BAud rate
  CommandString             BYTE "BOPRDHGAIPLONOPAQUSASUTEUNWEDNBA",0  'For command parsing
  ProtName                  BYTE "DMX    ","RENARD ","FUTURE1","FUTURE2",0 'Bank protocol names

VAR

'DHCP variables
  BYTE IP[4]                    'IP, initially 169.254.74.73, later as assigned by DHCP, or set static by Web page
  BYTE SubnetMask[4]            'Set by DHCP or forced if static
  BYTE GatewayIP[4]             'Set by DHCP or forced if static
  BYTE DHCP_Server[4]           'IP address of DHCP server
  BYTE DNS_Server[4]            'IP address of DNS server, set by DHCP, forced if static
  LONG DHCP_LeaseTime           'Duration of DHCP lease in seconds, no automatic renewal

'Get variables
  BYTE  Num2str[12]             'Convert numbers to a string here, 0 term
  BYTE  VarExists[10]           'Non-zero means that the string variable exists
  LONG  VarStr[10]              'Pointers to string variables passed in the URL when our web server connects

'Variables used by the spin code
  BYTE  Command[40]             'The entered command string from the web page gets put here
  BYTE  ThisNetwork             'Value showing how current network address was established:
  BYTE  CmdFnsh
  
'Variables used for Wiz routines
  BYTE  LEW                     'Last Eeprom Write, set 1-8 if last command was to write to the EEPROM
  BYTE  Socket3Mode             'E1.31/Web mode for 3rd socket
  BYTE  DestIP[4]               'Destination IP for UDP multicast packets
  BYTE  DestMAC[6]              'Destination MAC address for UDP multicast
  WORD  Universe[4]             'Universe number from packet
  WORD  SlotSize[4]             'E1.31 DMX number of channels 
  BYTE  WebState                'To handle web server since we can't do it all at one time
  BYTE  ParsedChars             'Character counter for the parser for commands
  WORD  ParamVal[10]            'Numeric parameters from the command line
  BYTE  WebSocket               'Socket number to be used by web server

'Variables used for timers - Used by ASM timer cog for PAR value (DO NOT MODIFY)
' Start of timers
  LONG  UpTime                  'Total system uptime
  LONG  E131idle[4]             'Time since last packet received
  LONG  WebIdle                 'Time since last connection established to the web server (not yet implemented)
  LONG  SpareUpTimers[6]        'Spare timers (not yet implemented)
  LONG  DHCPtimeleft            'Time remaining on DHCP lease before renew needed (not yet implemented)
  LONG  ServerTimer             'Timer for web server
  LONG  SpareDownTimers[2]      'Spare timers (not yet implemented)
  LONG  LedCode                 'Flash code for Red LED to be annunciated by aux cog (not yet implemented)
' End of timers             

'Variable used for Web Page stats
  LONG  StatPktRcv[4]           'For statistics, number of packets received for each universe

'Below are all of the parameters that are stored in the EEPROM.
'We allow for 8 pages of EEPROM params, numbered 0-7.
'At boot, we read them from EEPROM page 0 by default.

'Start of parameter table (DO NOT MODIFY) 
  WORD  Offset[4]               'Universe offset, 1-63999
  BYTE  MAC[6]                  'MAC address of this board. WIZ has no built-in MAC address.
  BYTE  Protocol[4]             '0=DMX , 1=REN, 2=FUTURE1, 3=FUTURE2
  BYTE  StaticIP[4]             'LAN info to use if IPmode=static
  BYTE  StaticGateway[4]        'Gateway IP for static IP mode
  BYTE  StaticSubnetMask[4]     'Subnet mask for static IP mode
  BYTE  StaticDNS[4]            'DNS address for static IP mode
  BYTE  Baudrate[4]             'Baud rate variable for Renard output
  BYTE  BootMode                'What we do at boot (not yet defined)
  BYTE  WebMode                 'Web server mode bit 0= if forced bit 1= after any reboot, 2= if no packets on socket 3
  BYTE  IPmode                  '0=Normal, DHCP if available, else use 169.254.74.73, 1=Use static only (partially implemented)
  BYTE  LOSmode                 'Loss of signal mode (not yet implemented)
' End of table

'Temporary storage buffers
  BYTE  WebpacketBuffer[WebpacketBufferLength] 'Packet buffer for web server
  LONG  DMXBuffer[513]        'DMX Output Data Buffers
  BYTE  PacketBuffer[E131PacketBufferLength]   'Received UDP packet goes here for checks
          
OBJ

  WIZ:          "W5100_SPI_Driver.spin"    'Wiznet SPI mode driver, from Obex                   
  EEPROM:       "Basic_I2C_Driver.spin"    'I2C driver for EEPROM access, from Obex (modified for our use) 
' Debug:        "FullDuplexSerial.spin"    'Serial Term for debugging, from Obex
  RR:           "RealRandom.spin"          'Random Number Generator, from Obex
  FMT:          "Numbers.spin"             'Misc numeric/string convert/format functions, from Obex
  DMXout[4]:    "jm_dmxout.spin"           'DMX output driver, from Obex (modified for our use)
  RENout[4]:    "Renard_Driver.spin"       'Renard ouput driver, by Gregory Bartlett (modified for our use)
  DHCP:         "DHCP"                     'DHCP object, taken from Jim's original E680 code. Original code by Roy Eltham
                
PUB Main | idx,Temp,ButtonCode,Socket,PacketSize 'Main Method - Start of Spin code, Main entry point

DIRA[GrnLED]:=_OUTPUT           'Set Red and Grn LED pins for output
DIRA[RedLED]:=_OUTPUT

'Debug.start(31, 30, 0, 57600    'Start the serial terminal object for debugging (Net yet implemented)   
'WaitCnt (Cnt+second*2)          '2 sec delay here to allow enabling of debug screen if used

FMT.init                        'Initialize the numbers formatting routine
ButtonCode:=BootButton          'Read the MODE button code from the pushbutton switch at boot time    

ReadParamPage(0)                'Read in parameter page 0 by default
Temp:=CheckParams               'and check it. Check code based on return code from routine

if  ButtonCode==8               'Check for restore to defaults on MODE Button Code #8
  LoadDefaults                  'If valid, reload defaults
  Temp:=ButtonCode              'and flag for Eeprom to be rewritten with defaults.
  ButtonCode:=0

if  ButtonCode==10              'Check for generate new MAC on MODE Button Code #10
  GenMAC                        'If valid, generate new random MAC (if no MAC IC installed)
  LoadDefaults                  'and reload defaults
  Temp:=ButtonCode              'then flag for Eeprom for all 8 pages to be rewritten with defaults.
  ButtonCode:=0

if (Temp>0)                     'If there were any values out of range, assume corrupt or
  repeat idx from 0 to 7        'uninitalized Eprom data and re-write all 8 parameter pages
     WriteParamPage(idx)
     reboot                     'and reboot

WIZ.StartSPI(_SCS, _SCLK, _MOSI, _MISO, _RESET)         'Initialize the wiz module
PauseMSec(1000)                                         'then 1 second delay

WIZ.WriteMACaddress(true,@MAC[0])'Write our MAC address to the Wiznet module 
PauseMSec(100)

                                'Now see what our IP mode is, if 0 we will default to 169.254.74.73 if no DHCP success, otherwise we use DHCP
                                'If mode 1 we will just use our static IP from the parameter block      
ThisNetwork:=IPmode             'initially set to what is stored in EEPROM, 0=DHCP, 1=Static but can change

if (WebMode&1)>0
     ServerTimer:=120           'Bring server up if WebMode bit 0 set

if ButtonCode >0
     ServerTimer:=999           'Force server up on any MODE Button Code

if ButtonCode==2
                                'A MODE button code of 2 forces IP mode to static for this boot
     ThisNetwork:=3             'show that it was forced static
     
if ButtonCode==3
                                'A MODE button code of 3 forces DHCP IP mode for this boot
     ThisNetwork:=2             'show that it was forced DHCP

if (ThisNetwork & 1)==1         'If this boot is static
    repeat idx from 0 to 3
       IP[idx]:=StaticIP[idx]
       GatewayIP[idx]:=StaticGateway[idx]
       SubnetMask[idx]:=StaticSubnetMask[idx]
       DNS_Server[idx]:=StaticDNS[idx]

if (ThisNetwork & 1)==0         'If this boot is DHCP
   DHCP.DHCPstart(DEFAULT_HOST_NAME)'Start cog and initialize our DHCP variables
   Temp:=$4a000000+(MAC[3]<<16)+(MAC[4]<<8)+MAC[5] 'build XID for DHCP
   idx:=DHCP.DoDHCP(0,Temp) 'try the DHCP
       if idx==0                'If DHCP successful
          DHCPTimeLeft:=DHCP_Leasetime   
       else
          SubnetMask[0]:=255    'If not successful, set the proper subnet mask for our 16-bit default address 255.255.0.0
          SubnetMask[1]:=255
          SubnetMask[2]:=0
          SubnetMask[3]:=0
          IP[0]:=169            'and set IP address to 169.254.74.73
          IP[1]:=254
          IP[2]:=74
          IP[3]:=73
          ThisNetwork:=ThisNetwork|4 'Show DHCP failed

   DHCP.stop                    'DHCP all done, now stop cog

repeat idx from 0 to ThisNetwork'Display current net status on the Grn LED
    FlashGrn
    WaitCnt(Cnt+second/8)       'Delay 1/8th second

                                'Now write the network values to the Wiz
WIZ.WriteGatewayAddress(true,@GatewayIP[0])
WIZ.WriteSubnetMask(true,@SubnetMask[0])
WIZ.WriteIPAddress(true,@IP[0])

DestIP[0]:=239                  'Set destination IP for Universe #1, modify later
Destip[1]:=255                  'add the rest later
                  
DestMAC[0]:=$01                 'Set destination MACc address for E1.31 prefix
DestMAC[1]:=$00
DestMAC[2]:=$5E 
DestMAC[3]:=$7f 

cognew (@auxcog,@uptime)        'Start aux cog for various timekeeping functions 

'Time to do a Web page
'
'If active output on 3rd socket, at bootup we come up in browser,
'wait up to 30 seconds for a page hit.  If no page hits after 30 secs
'we go into E1.31 mode on socket 3.  If there are hits, then we set
'the timeout timer to 120 seconds after the last hit, and it's reset for every hit.

'While in E1.31 mode, we look for a lack of E1.31 packets for xx seconds.
'If we see that, we pop back to browser mode with a fairly short timeout, if we get
'a hit we stay in browser mode, else we alternate 3 secs browser, 1 sec E1.31, until
'one or the other shows life.

'Single Web config page for now, will display everything, lan info, system config, stats, etc.
'get form for data editing, field name, value

  repeat idx from 0 to 3        'Set startup DMX slot size for cogs.
    SlotSize[idx]:=513

   ' Init TX Output Channel Bank cogs (Routines will change later when additional protocols are added)

  if Protocol[0]==1             'If protocol flag set for Renard, load Renard cog, else load DMX Cog
    DMXBuffer[0] := RENout[0].BuffrPtr
    DIRA[OUT1_Ctrl]:=1                                  'Set port for output
    OUTA[OUT1_Ctrl]:=1                                  'Set RS485 driver to output mode
    RENout[0].start(OUT1_Tx, 512, REN_Refresh, Baudrate[0])      ' Renard Output Bank 1
    RENout[0].flushbuf
  else 
    
     DMXBuffer[0] := DMXout[0].BuffrPtr
     DMXout[0].init(OUT1_Tx, OUT1_Ctrl, 0, 513, DMX_b2bt)' DMX Output Bank 1
     DMXout[0].flushbuf

  if Protocol[1]==1             'If protocol flag set for Renard, load Renard cog, else load DMX Cog
    DMXBuffer[1] := RENout[1].BuffrPtr
    DIRA[OUT2_Ctrl]:=1                                  'Set port for output
    OUTA[OUT2_Ctrl]:=1                                  'Set RS485 driver to output mode
    RENout[1].start(OUT2_Tx, 512, REN_Refresh, Baudrate[1])      ' Renard Output Bank 2
    RENout[1].flushbuf
  else 
    
     DMXBuffer[1] := DMXout[1].BuffrPtr
     DMXout[1].init(OUT2_Tx, OUT2_Ctrl, 0, 513, DMX_b2bt)' DMX Output Bank 2
     DMXout[1].flushbuf

  if Protocol[2]==1             'If protocol flag set for Renard, load Renard cog, else load DMX Cog
    DMXBuffer[2] := RENout[2].BuffrPtr
    DIRA[OUT3_Ctrl]:=1                                  'Set port for output
    OUTA[OUT3_Ctrl]:=1                                  'Set RS485 driver to output mode
    RENout[2].start(OUT3_Tx, 512, REN_Refresh, Baudrate[2])      ' Renard Output Bank 3
    RENout[2].flushbuf
  else 
    
     DMXBuffer[2] := DMXout[2].BuffrPtr
     DMXout[2].init(OUT3_Tx, OUT3_Ctrl, 0, 513, DMX_b2bt)' DMX Output Bank 3
     DMXout[2].flushbuf

  if Protocol[3]==1             'If protocol flag set for Renard, load Renard cog, else load DMX Cog
    DMXBuffer[3] := RENout[3].BuffrPtr
    DIRA[OUT4_Ctrl]:=1                                  'Set port for output
    OUTA[OUT4_Ctrl]:=1                                  'Set RS485 driver to output mode
    RENout[3].start(OUT4_Tx, 512, REN_Refresh, Baudrate[3])      ' Renard Output Bank 4
    RENout[3].flushbuf
  else 
    
     DMXBuffer[3] := DMXout[3].BuffrPtr
     DMXout[3].init(OUT4_Tx, OUT4_Ctrl, 0, 513, DMX_b2bt)' DMX Output Bank 4
     DMXout[3].flushbuf

 'Once we have started up cogs to drive the outputs, this cog can focus on reading E1.31 packets.
                                                       
repeat idx from 0 to 3          'Open all 4 Wiz sockets for E1.31 read.
  SocketReopen(idx)   
    
'TestMode:=0                    'Test pattern mode (not yet implemented)
'DebugDisplay:=0                'Turn on debug dislpay (not yet implemented)

repeat    'loop forever
  
  WebSocket:=3                                         'For now, web server will be on socket 3, maybe dynamic later #########
  repeat Socket from 0 to 3                            'Loop for every wiz socket

    if (Socket<>3) OR (ServerTimer==0)                 'Quit the repeat loop

       PacketSize:=WIZ.rxUDP (Socket,@PacketBuffer)    'Read incoming E1.31 packet

      if (PacketSize<133) or (PacketSize>648)          'Ignore if not proper packet length range for E1.31

      else                                             'Test for valid strings to qualify E1.31 packet

        if not ((strcomp(@RLPstr,@PacketBuffer[8]))and(strcomp(@DMPstr,@PacketBuffer[125])))

        else                                           'Checks OK, now process data             

           GrnOn                                       'Green LED on to indicate processing a valid packet

           Universe[Socket]:=PacketBuffer[121]*256+PacketBuffer[122] 'Save Universe #

           if SlotSize[Socket]<>PacketBuffer[131]*256+PacketBuffer[132] 'Check if current E1.31 DMX Slot count matches previous
              SlotSize[Socket]:=PacketBuffer[131]*256+PacketBuffer[132] 'If not, save new DMX Slot count
           
              If Protocol[Socket]==0                    'Check if DMX protocol for this bank
                DMXout[Socket].Slots(SlotSize[Socket])  'Dynamically change DMX channel count

              else
                                                                                  
                if Protocol[Socket]==1                    'Check if Renard protocol for this bank
                  RENout[Socket].Slots(SlotSize[Socket]-1)'Dynamically change Renard channel count


           if (PacketBuffer[133]==0)                    'If DMX start code is 0
              ++StatPktRcv[Socket]                      'Increment the packet received counter for stats
              E131idle[Socket]:=0                       'Reset idle timer, time since last packet received
              bytemove(DMXBuffer[Socket],@PacketBuffer[133],PacketSize-133) 'Copy E1.31 data slots 1-n to output buffer
            
           GrnOff                                  
    
    CheckWebServer                                     'Do the web server on socket 3

PUB ReadParamPage(page)         'Read one of the parameter pages from EEPROM into our parameter variable block

          if (page>7) or (page<0) 'Make sure page number is between 0-7, uses 512 bytes of EEPROM.
               return 99
          EEPROM.ReadPage(_EEPROMpin,_EEPROMsel,_EEPROMbase+(page*64),@Offset[0],8)  'Read the words first
          PauseMSec(10)

          EEPROM.ReadPage8bit(_MACROMpin,_MACROMsel,_MACROMbase,@MAC[0],6)'Get MAC address from Microchip MAC IC
          PauseMSec(10)
          if (MAC[0]<>$00) and (MAC[1]<>$04) and (MAC[2]<>$A3) 'if first 3 MAC bytes are not Microchip OUI from our MAC IC                  
            EEPROM.ReadPage(_EEPROMpin,_EEPROMsel,_EEPROMbase+(page*64)+9,@MAC,6)'then load our generated MAC from EEPROM

          PauseMSec(10)
          EEPROM.ReadPage(_EEPROMpin,_EEPROMsel,_EEPROMbase+(page*64)+15,@Protocol,28)'Now read the bytes
          PauseMSec(10)
          
PUB WriteParamPage(page)        'Write the current config params to a 64-byte EEPROM page

          if (page>7) or (page<0) 'Make sure page number is reasonable
               result:=99
               return
          result:=result|EEPROM.WritePage(_EEPROMpin,_EEPROMsel,_EEPROMbase+(page*64),@Offset[0],8)  'Write the words first
          PauseMSec(10)
          result:=result|EEPROM.WritePage(_EEPROMpin,_EEPROMsel,_EEPROMbase+(page*64)+9,@MAC,6)'Now write the bytes
          PauseMSec(10)
          result:=result|EEPROM.WritePage(_EEPROMpin,_EEPROMsel,_EEPROMbase+(page*64)+15,@Protocol,28)'Now write the bytes
          PauseMSec(10)
          'Non-zero result = EEPROM write error, 99 = Invalid page number 

PUB CheckParams | Bank, idx       'Check params read from EEPROM for validity, return result 1-255 according to combo of errors

      result:=0
      repeat Bank from 0 to 3                           'Check all bank-specific parameters

          if (Protocol[Bank]<0) or (Protocol[Bank]>MaxProtocols)
                Protocol[Bank]:=0
                result:=result|1

          if (offset[Bank]<1) or (offset[Bank]>63999)
                offset[Bank]:=Bank+1
                result:=result|2

          if (Baudrate[Bank]<0) or (Baudrate[Bank]>1)
                Baudrate[Bank]:=0
                result:=result|4

      if (BootMode>MaxBootMode)
                BootMode:=0
                result:=result|8
      if (LOSmode>MaxLosMode)
                LOSmode:=0
                result:=result|16
      if (WebMode>MaxWebMode)
                WebMode:=0
                result:=result|32
      if (IPmode>MaxIPmode)
                IPmode:=0       'Forces DHCP mode 
                result:=result|64

     if (MAC[0]==$00) and (MAC[1]==$04) and (MAC[2]==$A3) 'if first 3 MAC bytes are Microchip OUI                  

       return                                             'then return             

     else                                                 'else check for our standard MAC

      if (MAC[0]<>$4A) or (MAC[1]<>$49) or (MAC[2]<>$4D)  'if first 3 MAC bytes aren't our generated prefix
       
         GenMAC                 'Generate new random MAC
         
         result:=result|128     'flag bad MAC
       
         LoadDefaults           'and reload all defaults
     
PUB LoadDefaults | Temp         'Load default parameters

         WebMode:=3             'Web server on
         IPmode:=1              'Static IP mode
         BootMode:=0            'Boot mode - Webserver on at boot
         LOSmode:=0             'Loss Of Signal mode (not yet implemented) 

         StaticIP[0]:=192       'Static IP address
         StaticIP[1]:=168
         StaticIP[2]:=1
         StaticIP[3]:=206

         StaticSubnetMask[0]:=255 'Static subnet
         StaticSubnetMask[1]:=255
         StaticSubnetMask[2]:=255
         StaticSubnetMask[3]:=0

         StaticGateway[0]:=192    'Static Gateway address
         StaticGateway[1]:=168
         StaticGateway[2]:=1
         StaticGateway[3]:=1

         StaticDNS[0]:=192        'Static DNS address
         StaticDNS[1]:=168
         StaticDNS[2]:=1
         StaticDNS[3]:=1

         repeat Temp from 0 to 3      'Reset all banks to DMX protocol & Banks to DMX Universe 1-4

            Offset[Temp]:=Temp+1
            Protocol[Temp]:=0
            Baudrate[Temp]:=0         'Reset Renard bank baud rates to 57600

PUB GenMAC                      'Generates new random MAC

  MAC[0]:=$4A                   'Initialize the first 3 bytes to our standard prefix
  MAC[1]:=$49
  MAC[2]:=$4D
  RR.start                      'Start the random number generator
  MAC[3]:=RR.random             'Generate the last 3 bytes from the RNG
  MAC[4]:=RR.random
  MAC[5]:=RR.random
  RR.stop                       'Now stop the cog                       
  
PUB SocketReopen(skt)           'New method, will close, then reopen socket[socket], use when universe changes on the fly 

      WIZ.socketclose(skt)                          'First close Wiz socket
      PauseMSec(10)                                 'Pause for 10ms(was 200msec here) #####
      DestIP[2]:=offset[skt]>>8                     'Make proper destination IP address by changing 2 LS bytes to be universe #
      DestIP[3]:=offset[skt] & $FF                       
      WIZ.SocketOpen(skt,$82,0,5568,@destIP[0])     'Open the socket for multicast read

PRI CheckWebServer | Temp       'Bring Web server up or down dynamically

      if (Socket3Mode==0) AND ((WebMode&2)>0) AND (E131idle[3]>12)  'If need to bring up server due to E131 timeout
           ServerTimer:=3                           'set it to timeout at 3 seconds, but if we get a hit we'll bump this up
           
      if (Socket3Mode>0) AND (ServerTimer==0)       'If we are stopping the server
           SocketReopen(3)                          'reopen socket for e1.31 reception
           Socket3Mode:=0                           'show socket as back in e1.31 mode
           E131idle[3]:=10                          'reset the socket e1.31 packet loss timer 
                               
      if (ServerTimer>0)
          RedOn
         if (Socket3Mode==0)                        'We are starting the server
              WebState:=0                           'Force socket to be opened for web server
              Socket3Mode:=1                        'Show socket mode now for web
      else
          RedOff
          return

 
    if (WebState<5) or (WebState>150)
         
case WebState    'Now we process commands based on value of WebState variable.   

  0:
     
     PauseMSec (10)                    '#### test to elim syn/rst hangs
     WIZ.SocketOpen(WebSocket, Wiz#_TCPPROTO, LocalSocket, DestSocket, @DestIP[0])
     PauseMSec (10)                    'best value? need more testing with this
     WIZ.SocketTCPlisten(WebSocket)
     ++WebState
     
  1:
    
    if WIZ.SocketTCPestablished(WebSocket)==0 'if no connection now
      Temp:=DHCP.readstatus(WebSocket)

      if (Temp<>20) AND (Temp<>22) AND (Temp<>23)
           WebState:=199
           return

    else   
        bytefill(@WebpacketBuffer,0,WebpacketBufferLength)
        PauseMSec(3)
        WIZ.rxTCP(WebSocket,@WebpacketBuffer)                       'Read in TCP packet if one has been received

        if WebpacketBuffer[0] <> "G" ' Assume a GET request         'If 1st character of packet is 'G' we assume it's a Get request
           return

        else
          if (ServerTimer<300)
                ServerTimer:=300
          WebState:=2
          return

  2:    
           ParseURL                                          'Parse incoming URL string

           if byte[VarStr[1]] <> 0                           'if a command was entered
                WebState:=5                                  'this will be our command processing
                return

           WebState:=100                                     'skip command processing, go to page render
           return

  5:                                                         'command processing here
        command:=(MatchString(varstr[1],@commandstring[0],2))'try to match 1st 2 characters of a defined command word

        if command==0
           WebState:=100                                     'no valid command, go to page render
           return
       
        WebState:=8+command*2
        return

  10:     'BO = Boot command, dual purpose, sets boot mode, or if 999 is entered, reboots the board
        GetParams(0)                                        'Get boot mode number

        if (ParamVal[0]=<maxbootmode)                       'See if in range for legal bootmode command
             bootmode:=ParamVal[0]                          'if so, save it

        if ParamVal[0]==999                                 'else, see if it's a system reboot command
             reboot

        WebState:=100
        return

  12:     'PR = Bank protocol
        GetParams(1)                                        'Get the bank protocol type

        if (ParamVal[1]=>1) AND (ParamVal[1]=<MaxProtocols) 'If legitimate protocol
               Protocol[ParamVal[0]-1]:=ParamVal[1]-1       'then set protocol type
               
        WebState:=100
        return

  14:     'DH = DHCP command
        
        GetParams(0)                                        'Get DHCP mode number 0/Static 1/DHCP?

        if ParamVal[0]<2
            ipmode:=1-ParamVal[0]
            
        WebState:=100
        return

  16:     'GA = Static Gateway command
        GetDQ(@StaticGateway[0])
        WebState:=100
        return

  18:     'IP = Static IP command
        GetDQ(@StaticIP[0])
        WebState:=100
        return

  20:     'LO = Loads a ParamPage(n) from EEPROM command
        GetParams(0)     'Get page number from 0 to 7

        if ParamVal[0]<8
           ReadParamPage(ParamVal[0])
        WebState:=100
        return

  22:     'NO = No Signal Mode (command  not yet implemented)
        GetParams(0)     'get page number

        if ParamVal[0]<256
          LOSmode:=ParamVal[0]

        WebState:=100
        return

  24:     'PA = Password command (not yet implemented)
        WebState:=100
        return

  26:     'QU = Quit command
        if GetParams(0)==0

            ServerTimer:=10         'Set Server to timeout in 10 seconds
        else
            ServerTimer:=10+ParamVal[0]

        WebState:=100
        return

  28:     'SA = Save to EEPROM command - WriteParamPage(n)
        GetParams(0)     'Get page number

        if ParamVal[0]<8
            'if (LEW<>(ParamVal[0]+1))   ###not working properly, a page load neads to clear LEW fix later #######
                 WriteParamPage(ParamVal[0])  'we won't do the write if the last command was a write to this same page

            'if ParamVal[0]==0
               'parampage2par   're-write the params passed to the asm cogs whenever we write EEPROM page 0

        WebState:=101
        LEW:=ParamVal[0]+1      'last command was an EEPROM write, keep track to avoid repeat writes with F5 brouser refresh
        return

  30:     'SU = Static Subnet Mask command
        GetDQS(@StaticSubnetMask[0])

        WebState:=100
        return

  32:     'TE = Test Pattern Command (not yet implemented)
        GetParams(0)     'Get page number

        'if ParamVal[0]=<MAXtp
        '  TPmode:=ParamVal[0]

        WebState:=100
        return

  34:     'UN = Universe command
        GetParams(1)   'get the socket number 1-4 and the universe number for it

        if (ParamVal[0]>0) AND (ParamVal[0]<5)    'if legitimate bank number

             if ParamVal[1]< 64000
                 offset[ParamVal[0]-1]:=ParamVal[1]  'offset=universe number

                 if (ParamVal[0]<4)               'don't lose/reopen socket 3 bacause it's running the web server
                      SocketReopen(ParamVal[0]-1) 're-open a socket if the universe offset changes

        WebState:=100
        return

  36:      'WE = Web Mode command
        GetParams(0)

        if (ParamVal[0]<4)
            WebMode:=ParamVal[0]

        WebState:=100
        return

  38:     'DN = Static DNS command
        GetDQ(@StaticDNS[0])

        WebState:=100
        return '

  40:     'BA = Baud rate - Renard options
        GetParams(1)                                        'Get the bank baud rate

        if (ParamVal[1]=>1) AND (ParamVal[1]=<2)            'If legitimate value
               Baudrate[ParamVal[0]-1]:=ParamVal[1]-1       'then set baud rate
               
        WebState:=100
        return


  10..98:   'Catch any bad values, shouldn't get here
      WebState:=100
      return         
  

  100: LEW:=0            'Last command wasn't a save command
       ++WebState
       return

  101:
      'Send Web Page - This is the start of our Web page  
      StringSend(WebSocket,String("HTTP/1.1 200 OK", CR, LF))
      StringSend(WebSocket,String("Server: E1.31 to DMX Bridge Configuration Page", CR)) 'header
      StringSend(WebSocket,String("Connection: close", CR))
      StringSend(WebSocket,String("Content-Type: text/html", CR, LF))
      StringSend(WebSocket,String(CR, LF))
       
      StringSend(WebSocket,String("<HTML>", CR))
      StringSend(WebSocket,String("<HEAD>", CR))
      StringSend(WebSocket,String("<TITLE>"))
      StringSend(WebSocket,String("E1.31 to DMX Bridge Configuration Page"))
      StringSend(WebSocket,String("</TITLE>", CR))
      StringSend(WebSocket,String("</HEAD>", CR))                                                       

      StringSend(WebSocket,String("<BODY onload = 'focusIt()'>", CR)) 'was document.forms.MyForms.
      StringSend(WebSocket,String("<H2>"))         'H2=Medium size font
      StringSend(WebSocket,String("<CENTER>E1.31 to DMX Bridge Configuration Page</CENTER>"))
      StringSend(WebSocket,String("</H2>", CR))

      StringSend(WebSocket,String("<HR>", CR))    'Makes a horizontal line
      WebState:=103
      return
      
  103:
       Stringsend(WebSocket,String("Current Network Mode is: "))
      if (ThisNetwork & 2)>0      'If IPmode was forced at bootup
          StringSend(WebSocket,String("FORCED"))
      else
          StringSend(WebSocket,String("Normal"))    

      if (ThisNetwork & 1)==0
          StringSend(WebSocket,String(" DHCP  "))
      else
          StringSend(WebSocket,String(" Static"))
      if (ThisNetwork & 4)>0
          StringSend(WebSocket,String("FAILED"))
      StringSend(WebSocket,String("<PRE> IP Address: "))
      WebState:=105
      return

  105:
      DQSS (WebSocket,@IP[0])
      StringSend(WebSocket,String("   Subnet Mask: "))
      DQSS (WebSocket,@subnetmask[0])
      WebState:=109
      return

  109:
      StringSend (WebSocket,String(CR,"    Gateway: "))
      DQSS (WebSocket,@gatewayip[0])
      StringSend (WebSocket,String("    DNS Server: "))
      DQSS (WebSocket,@dns_server[0])
      StringSend(WebSocket,String("<BR>"))
      WebState:=113
      return     
      
  113:
      if (ThisNetwork&1)==0      'If DHCP
         StringSend(WebSocket,String("DHCP Server: "))
         DQSS (WebSocket,@dhcp_server[0])
      StringSend(WebSocket,String("MAC Address: "))
      MACSS (WebSocket,@MAC[0])
      WebState:=117
      return

  117:
      if (ThisNetwork&1)==0      'If DHCP
         StringSend (WebSocket,String(" DHCP  Lease Time Remaining: "))
         TimeSS (WebSocket,DHCPTimeLeft)
      StringSend(WebSocket,String("<BR>"))
     
              
      WebState:=121
      return

  121:
      StringSend(WebSocket,String("  System Up-Time: "))
      TimeSS (WebSocket,uptime)
      WebState:=125
      return

  125:
      StringSend(WebSocket,String("     Firmware Rev: "))
      n2s (MajorVersion,1)
      StringSend(WebSocket,String("."))
      n2s (MinorVersion,2)
      StringSend(WebSocket,String("   Web Server Timeout in: "))
      if ServerTimer<1000
            n2s (ServerTimer,3)
            StringSend(WebSocket,String(" Seconds"))
      else
          StringSend(WebSocket,String("---"))
      StringSend(WebSocket,String("<BR></PRE>"))
      StringSend(WebSocket,String("<HR>"))     'Makes a horizontal line
      StringSend(WebSocket,String("<PRE>E1.31 Packet Statistics:"))
      StringSend(WebSocket,String("  Socket S1    Socket S2    Socket S3    Socket S4<BR><BR>"))
      WebState:=129                                                                          
      return
           
  129:
      StringSend(WebSocket,String(" Universe Number     "))
      repeat Temp from 0 to 3
            StringSend(WebSocket,FMT.tostr(offset[Temp],$2298A))
            StringSend(WebSocket,String(" "))
      StringSend(WebSocket,String("<BR>"))     
      StringSend(WebSocket,String(" Packets Received    "))
      WebState:=133
      return
      
  133:    
      repeat Temp from 0 to 3
            StringSend(WebSocket,FMT.tostr(StatPktRcv[Temp],$2298A))
            StringSend(WebSocket,String(" "))
      StringSend(WebSocket,String("<BR>"))      

      WebState:=141
      return

  141:

   StringSend(WebSocket,String(" DMX Channel Count   "))
     repeat Temp from 0 to 3
           StringSend(WebSocket,FMT.tostr(SlotSize[Temp]-1,$2298A))
           StringSend(WebSocket,String(" "))
 
     WebState:=145
     return
  
  145:
      StringSend(WebSocket,String("<HR>"))    'Makes a horizontal line
      StringSend(WebSocket,String("Global Configuration:<BR>"))
      StringSend(WebSocket,String("<PRE>Static Network Information:",CR,"IP Address: "))
      WebState:=149
      return

  149:
      DQSS (WebSocket,@staticIP[0])
      StringSend(WebSocket,String("   Subnet Mask: "))
      WebState:=153
      return
      
  153:
      DQSS (WebSocket,@StaticSubnetMask[0])
      StringSend (WebSocket,String(CR,"   Gateway: "))
      DQSS (WebSocket,@StaticGateway[0])          
      WebState:=157
      return

  157:
      StringSend (WebSocket,String("    DNS Server: "))
      DQSS (WebSocket,@StaticDNS[0])
      WebState:=161
      return

  161:       
      StringSend(WebSocket,String("<BR>",CR))
      StringSend(WebSocket,String("Default IP Mode: "))
      if ipmode==0
            StringSend(WebSocket,String("DHCP  "))
      else
            StringSend(WebSocket,String("STATIC"))
      StringSend(WebSocket,String("  Web Server Mode: "))
      n2s (WebMode,1)

     ' ** NOT YET IMPLEMENTED **
      'StringSend(WebSocket,String("  No Data Timeout: "))
      'n2s (LOSmode,3)
      'StringSend(WebSocket,String("  Test Pattern: "))
      'n2s (TPmode,1)

     StringSend(WebSocket,String("  Bootup Initial Mode: "))
      n2s (bootmode,1)
      StringSend(WebSocket,String("<BR>"))
        
      StringSend(WebSocket,String("<HR></PRE>"))    'makes a horizontal line
      StringSend(WebSocket,String("Output Configuration:<BR>"))
      StringSend(WebSocket,String("<PRE>Output#      Universe     Protocol<BR>"))
      WebState:=165
      return
      
  165,167,169,171: 'Do half of a line per pass for this range
        Temp:=(WebState-165)/2 'Temp will be 0-3 = Bank#
        StringSend(WebSocket,String("   "))
        n2s (Temp+1,1)

        StringSend(WebSocket,String("   "))
        StringSend(WebSocket,FMT.tostr(offset[Temp],$2298A))
        StringSend(WebSocket,String("        "))

        c2s (Protocol[Temp])
        if Protocol[Temp] == 1                          'Check if bank protocol is Renard
          if Baudrate[Temp] == 0                        'If so then display bank baud rate 
            StringSend(WebSocket,String("@ 57,600 Baud "))
            RENout[Temp].chbaudrate(0)
          else
            StringSend(WebSocket,String("@ 115,200 Baud"))
            RENout[Temp].chbaudrate(1)
        
        ++WebState
        return

  166,168,170,172:   'Do 2nd half of a line of Bank info for this range
        Temp:=(WebState-166)/2       'Will be 0-1-2-3
        StringSend(WebSocket,String("    "))
        StringSend(WebSocket,String("<BR>"))
        ++WebState
        return  
          
  173:

      StringSend(WebSocket,String(CR,"NOTE: Protocol changes require SAVE 0 and BOOT 999 to take effect",CR))  
      StringSend(WebSocket,String("<BR><FORM NAME='MYFORM' METHOD='GET'>"))  
      StringSend(WebSocket,String("Command: <INPUT NAME='1' SIZE='30'><BR>"))
      StringSend(WebSocket,String("</FORM>",CR))
      StringSend(WebSocket,String("<script language='javascript' type='text/javascript'>",CR ))
      StringSend(WebSocket,String("//window.alert('hello');"))
      StringSend(WebSocket,String("function focusIt()",CR))
      StringSend(WebSocket,String("{",CR))
      StringSend(WebSocket,String("var mytext = document.getElementById('1');",CR))
      StringSend(WebSocket,String("mytext.focus();",CR))
      StringSend(WebSocket,String("}",CR))
      StringSend(WebSocket,String("</script>",CR))
      WebState:=175
      return
 
  175:

      StringSend(WebSocket,String("</BODY>",CR,"</HTML>",CR,CR,LF))
      WebState:=199
      return

  99,199:
      PauseMSec (10)                                     '#### added trying to prevent syn/rst hangs  jjj
      WIZ.SocketTCPdisconnect(WebSocket)
      PauseMSec (10)                                     '#### see above
      WIZ.SocketClose(WebSocket)
      WebState:=0
      return

PRI DQSS(_socket, _dataptr) | dig,dq                    'Dotted quad string function
   'Send a dotted quad string
    repeat dig from 0 to 3
       bytemove (@dq,_dataptr,4)
       n2s(dq.byte[dig],3)
       if (dig == 3)
            quit
       StringSend(_socket,string("."))

PRI MACSS(_socket, _dataptr) | dig,dq
   'Send a MAC address formatted string
       bytemove (@dq,_dataptr,4)  'get first 4 bytes
       repeat dig from 0 to 3
              n2hs(dq.byte[dig])
             
              StringSend(_socket,string(":"))
       bytemove (@dq,_dataptr+4,2)  'get last 2 bytes
       n2hs(dq.byte[0])
       StringSend(_socket,string(":"))
       n2hs(dq.byte[1])                                
    
PRI TimeSS(_socket, _seconds) | dq                      'Time string
       dq:=_seconds/3600                                
       n2s(dq,4)                                        'Make as a 4-digit string
       StringSend(_socket,string($3A))
       _seconds:=_seconds-dq*3600                                        'minutes
       dq:=_seconds/60
       n2s(dq,2)                                        'Make as a 4-digit string
       StringSend(_socket,string($3A))
       _seconds:=_seconds-dq*60
       n2s(_seconds,2)

PRI StringSend(_socket, _dataPtr)
  if byte[_dataPtr] <> 0
    WIZ.txTCP(_socket, _dataPtr, strsize(_dataPtr))

PRI ParseURL | char, InPlace, OutPlace

  InPlace := 4 ' skip "GET "

  repeat char from 0 to 9                               'Reset all of our possible result variables
    VarExists[char]:= 0
    VarStr[char]:= 0

  repeat until (char == "?") or (char == " ")           'Step thru packet until ? or space, that will
    char:=WebpacketBuffer[InPlace++]                    'be just past the end of the command name

  WebpacketBuffer[InPlace-1]:=0                         'String terminator for command string 
  OutPlace := InPlace
                                      
  if char == "?" 'spcl chars use sequence %nn, comma is %2C  &=%26 % followed by ascii code of char
    repeat until char == " "
      char := WebpacketBuffer[InPlace]
      if (char =< "9") and (char => "0") 
        char := char - 48
        VarExists[char]:=1
        InPlace+=2 ' skip "x="
        VarStr[char] := @WebpacketBuffer[OutPlace]
        char := WebpacketBuffer[InPlace]
        if (char <> "&") and (char <> " ")
          WebpacketBuffer[OutPlace++]:=char
      repeat until (char == "&") or (char == " ")
        InPlace++
        char:=WebpacketBuffer[InPlace]
        if (char <> "&") and (char <> " ")
          if char == "%"
            InPlace++
            char:=WebpacketBuffer[InPlace++] - 48
            if char > 9
              char-=7
            char:=char << 4
            WebpacketBuffer[OutPlace]:=char
            char := WebpacketBuffer[InPlace] - 48
            if char > 9
              char-=7
            WebpacketBuffer[OutPlace++]+=char
            char:="x" 
          else
            if char=="+"
              WebpacketBuffer[OutPlace++] := " "
            else
              WebpacketBuffer[OutPlace++] := char
      WebpacketBuffer[OutPlace++]:= 0  ' String terminator
      InPlace++  

  'these methods are used for parsing the command line
PUB MatchString(string1,string2,CharsToMatch) | j,k,l
    
 ParsedChars:=0
 if CharsToMatch==0
     return                                             'CharsToMatch must be at least 1
 repeat k from 1 to 99                                  'this will be our result = which seg of string matched, 1-n
      if byte[string2]==0                               'if at end of string2
          return 0                                      'no match  if here
      l:=0                                              'set count of matched characters to 0
      repeat j from 1 to CharsToMatch                   'this loop is the char by char test
          if (byte [string1][j-1]&$DF)<>byte[string2][j-1]   'if not a match
             quit                                       'quit this repeat
          else
            ++l                                         'if a match, incr match counter
                                                        'when we finish that repeat, l is # of characters that matched
      string2:=string2+CharsToMatch                     'adv string 2 pointer to next segment of table string
      if l==CharsToMatch                                
         ParsedChars:=CharsToMatch                      'true if we have a full match, show that we processed CharsToMatch chars of the string
         return k     'successful match
 return 0 'shouldn't ever get here  

PUB NextParam(string1) |j,k                             'next numeric value in the command string 
       k:=0                                             'if digits found count, start at 0
       cmdfnsh:=0                                       'we'll set this 1 if we find end of string
       result:=0                                        
       repeat ParsedChars from 1 to 255                 'step through characters
          j:=byte[string1][ParsedChars-1]               'ascii code of next char
          if j==0
             cmdfnsh:=1                                 'mark that end of string found 
          if ((j<48) OR (j>57))                         'if not a digit
               if k>0                                   'if we've found some digits
                  return                                'we're done
               if j==0                                  'if end of string
                  if k==0                               'and we've found no digits
                    result:=-1                          'set result= -1 if no digits found
                  return
               next
          result:=result*10+j-48                        'if it's a digit, mult prior result by 10 and add value of this digit
          k:=k+1                                        'also bump count of digits found
                     
PUB GetParams(j) | k            'j is highest param# to get, 0=1 param, 1=2 params, etc
                                'returns actual number of numeric params found in result
     repeat k from 0 to 9
           ParamVal[k]:=0

     repeat k from 0 to j  'we may not always want to get every param if we're going to have
                           'to find an alpha param
           ParamVal[k]:=nextparam(varstr[1])   'get the next numeric parameter
           if ParamVal[k]==65535
              quit
           varstr[1]:=varstr[1]+ParsedChars
           result:=k+1
           if cmdfnsh==1
              quit
          
PUB GetDQS(j) | q,r,s           'Get a dotted quad in vaild subnet mask format
                                'May be entered as number of bits or as a dotted quad eg 255.255.255.0
       q:=GetParams(3)          'get up to 4 params 
       if q==1                  'if only 1 entered, assume it's a subnet mask bit size
           if (ParamVal[0]>0) and (ParamVal[0]<32) 'if it's a legitimate value for subnet mask
                s:=0
                repeat r from 31 to 32-ParamVal[0] 'if param=24, we set 24 highest bits, 31->8 
                     s:=s| (|<r)
                repeat r from 0 to 3
                     byte[j][r]:=s.byte[3-r]

       if q==4                      'if 4 params were entered, they are entering it in form of a dotted quad
            s.byte[0]:=ParamVal[3]    'build a long in s
            s.byte[1]:=ParamVal[2]
            s.byte[2]:=ParamVal[1]
            s.byte[3]:=ParamVal[0]
            q:=$80000000                   'ms bit only set
            repeat r from 0 to 31          'test for a match against every valid value
               if s==q                     'if we match
                    repeat r from 0 to 3   'put result in j and done
                       byte[j][r]:=s.byte[3-r]
                    return
               else
                   q:=q~>1          'else shift q right 1 bit, keeping sign, to test next value

PUB GetDQ(j) | k                'Read 4 numeric params, verify proper range for dotted quad, stuff them in
       GetParams (3)            'Read 4 params
       repeat k from 0 to 3
          if (ParamVal[k]==-1) or (ParamVal[k]>255)
             return
       repeat k from 0 to 3
          byte[j][k]:=ParamVal[k]

PUB FlashRed                    'Red LED Flash
          RedOn
          WaitCnt(Cnt+second/6)
          RedOff
          WaitCnt(Cnt+second/6)

PUB FlashGrn                    'Grn LED Flash
          GrnOn
          WaitCnt(Cnt+second/6)
          GrnOff
          WaitCnt(Cnt+second/6)

PUB RedOn                       'Red LED on
          OUTA[RedLED]:=1

PUB RedOff                      'Red LED off
          OUTA[RedLED]:=0

PUB GrnOn                       'Grn LED on
          OUTA[GrnLED]:=1

PUB GrnOff                      'Grn LED off
          OUTA[GrnLED]:=0

PUB Button                      'Button is TRUE if button pressed
          if INA[ProgPin]
             return
          result:=1

PUB BootButton                  'Check program button at boot, and return a value depending on
                                'if and how long it's held down for.
          result:=0             'Means button wasn't pressed, or wasn't released at the proper time so do nothing
          WaitCnt(Cnt+second*2) '2 second delay to give time to press the MODE button
          if Button==0
               return           'If button isn't pressed we're done
          repeat result from 1 to 10  'allow up to 10 boot switch operations
            RedOn
            GrnOn
            WaitCnt(Cnt+second/2)'Red LED on for 1/2 sec
            RedOff
            GrnOff
            WaitCnt(Cnt+second) 'Delay 1 more sec
            if button==0
                return          'If button released, return with code
          result:=0             'held down too long, return as if never pressed

PRI PauseMSec(Duration)         'Pause x milliseconds, used in multiple places
'  Pause execution for specified milliseconds.
'  This routine is based on the set clock frequency.
'  
'  params:  Duration = number of milliseconds to delay                                                                                               
'  return:  none
  WaitCnt(((clkfreq / 1_000 * Duration - 3932) #> 381) + Cnt)

PUB N2S(value,digits)| i,j      'Number convert to string in Num2str, pos only, left-space padded
   Num2str[digits]:=0
   i:=1
   if (digits>1)
     repeat j from 1 to (digits-1)
       i:=i*10
   repeat j from 0 to (digits-1)
      Num2str[j]:=(value/i)+$30
      value:=value-(value/i)*i
      i:=i/10
   StringSend(WebSocket,@Num2str[0])     

PUB N2HS(value) | i,j,k         'Convert number to hex for displaying MAC addresses
    Num2str[2]:=0
    k:=value/16+$30
    if k>$39
         k:=k+7
    Num2str[0]:=k
    k:=value//16+$30
    if k>$39
         k:=k+7
    Num2str[1]:=k
    StringSend(WebSocket,@Num2str[0]) 

PUB C2S(value) | i              'Convert output protocol code to 4-char descriptor string
    repeat i from 0 to 6
       Num2str[i]:=ProtName[value*7+i]
    Num2str[7]:=0
    StringSend(WebSocket,@Num2str[0]) 

PUB R2S(value) | i,j            'Converts a reverse flags word into a 4 char y/n string, y if 1, n if 0
     repeat i from 0 to 3
        Num2str[i]:="N"
        if (value & 1) ==1
           Num2str[i]:="Y"
        value:=value>>1
     Num2str[4]:=0
     StringSend(WebSocket,@Num2str[0]) 

DAT
'***************************************************************************************
                        org     0     
AuxCog
'----------------------------------------------------------------------------------
'this cog will handle timer functions
'---------------------------------------------------------------------------------- 

                        mov     jtimer,Cnt              'initialize 1 sec timer
                        add     jtimer,jsecond          
timeloop                waitcnt jtimer,jsecond          'synchronized 1 sec timer '
                        mov     timeraddress,PAR        'address of my long for total uptime in seconds, the 1st timer
                        mov     tlindex,#12             'number of uptimers
uptimerloop             rdlong  timervalue,timeraddress 'read the next timer
                        add     timervalue,#1           'add 1 second
                        wrlong  timervalue,timeraddress
                        add     timeraddress,#4         'adv timer address pointer to next timer
                        djnz    tlindex, #uptimerloop
                        mov     tlindex,#4              'number of downtimers
downtimerloop           rdlong  timervalue,timeraddress WZ'read the next timer
                IF_Z    jmp     #skipifzero
                        sub     timervalue,#1 'add 1 second
                        wrlong  timervalue,timeraddress
skipifzero              add     timeraddress,#4         'adv timer address pointer to next timer
                        djnz    tlindex, #downtimerloop
                        jmp     #timeloop

jsecond         long      80000000
timeraddress    res       1
timervalue      res       1
tlindex         res       1
jtimer          res       1

DAT
{{
+------------------------------------------------------------------------------------------------------------------------------+
                                                   TERMS OF USE: MIT License                                                  
+------------------------------------------------------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:                                                                   
                                                                                                                              
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
                                                                                                                              
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
+------------------------------------------------------------------------------------------------------------------------------+
}} 